__version__ = "0.3.1"

from importlib.util import find_spec

from .utils import *  # NOQA
from .hatch_dict.hatch_dict import *  # NOQA

from .extras.decorators.decorators import *  # NOQA
from .extras.decorators.mlflow_logger import *  # NOQA
from .extras.ops.argparse_ops import *  # NOQA

if find_spec("kedro"):
    from .pipeline.pipeline import *  # NOQA
    from .pipeline.sub_pipeline import *  # NOQA
    from .framework.configure import *  # NOQA
    from .framework.context.flexible_context import *  # NOQA
    from .extras.hooks.add_transformers import *  # NOQA
    from .extras.hooks.mlflow.mlflow_basic_logger import *  # NOQA
    from .extras.hooks.mlflow.mlflow_artifacts_logger import *  # NOQA
    from .extras.hooks.mlflow.mlflow_time_logger import *  # NOQA
    from .extras.hooks.mlflow.mlflow_outputs_logger import *  # NOQA
    from .extras.transformers.mlflow.mlflow_io_time_logger import *  # NOQA

if find_spec("pandas"):
    from .extras.datasets.pandas.csv_local import *  # NOQA
    from .extras.datasets.pandas.efficient_csv_local import *  # NOQA
    from .extras.datasets.pandas.pandas_cat_matrix import *  # NOQA
    from .extras.datasets.pandas.pandas_describe import *  # NOQA
    from .extras.datasets.pandas.histgram import *  # NOQA

if find_spec("pandas_profiling"):
    from .extras.datasets.pandas_profiling.pandas_profiling import *  # NOQA

if find_spec("PIL"):
    from .extras.datasets.pillow.images_dataset import *  # NOQA

if find_spec("seaborn"):
    from .extras.datasets.seaborn.seaborn_pairplot import *  # NOQA

if find_spec("torchvision"):
    from .extras.datasets.torchvision.iterable_images_dataset import *  # NOQA

if find_spec("cv2"):
    from .extras.datasets.opencv.images_dataset import *  # NOQA

if find_spec("requests"):
    from .extras.datasets.requests.api_dataset import *  # NOQA

if find_spec("httpx"):
    from .extras.datasets.httpx.async_api_dataset import *  # NOQA

if find_spec("numpy"):
    from .extras.ops.numpy_ops import *  # NOQA

if find_spec("pandas"):
    from .extras.ops.pandas_ops import *  # NOQA
    from .extras.decorators.pandas_decorators import *  # NOQA

if find_spec("torch"):
    from .extras.ops.pytorch_ops import *  # NOQA

if find_spec("ignite"):
    from .extras.ops.ignite.declaratives.declarative_trainer import *  # NOQA

if find_spec("shap"):
    from .extras.ops.shap_ops import *  # NOQA

if find_spec("sklearn"):
    from .extras.ops.sklearn_ops import *  # NOQA

if find_spec("allennlp"):
    from .extras.ops.allennlp_ops import *  # NOQA

if find_spec("cv2"):
    from .extras.ops.opencv_ops import *  # NOQA

if find_spec("skimage"):
    from .extras.ops.skimage_ops import *  # NOQA

if find_spec("memory_profiler"):
    from .extras.decorators.memory_profiler import *  # NOQA

if find_spec("pynvml") or find_spec("py3nvml"):
    from .extras.decorators.nvml_profiler import *  # NOQA



from pipelinex.utils import DictToDict
from skimage import segmentation


class SkimageSegmentationDictToDict(DictToDict):
    module = segmentation


class SkimageSegmentationFelzenszwalb(SkimageSegmentationDictToDict):
    fn = "felzenszwalb"


class SkimageSegmentationSlic(SkimageSegmentationDictToDict):
    fn = "slic"


class SkimageSegmentationQuickshift(SkimageSegmentationDictToDict):
    fn = "quickshift"


class SkimageSegmentationWatershed(SkimageSegmentationDictToDict):
    fn = "watershed"


class SkimageMarkBoundaries(SkimageSegmentationDictToDict):
    fn = "mark_boundaries"



import copy
from pathlib import Path
from typing import Any, Dict, Union
from PIL import Image
import logging

import numpy as np
from ..core import AbstractVersionedDataSet, DataSetError, Version

from ...ops.numpy_ops import to_channel_first_arr, to_channel_last_arr, ReverseChannel

log = logging.getLogger(__name__)


class ImagesLocalDataSet(AbstractVersionedDataSet):
    """ Loads/saves a dict of numpy 3-D or 2-D arrays from/to a folder containing images.
    Works like ``kedro.extras.datasets.pillow.ImageDataSet`` and
    ``kedro.io.PartitionedDataSet`` with conversion between numpy arrays and Pillow images.
    """

    def __init__(
        self,
        path: str,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = {"suffix": ".jpg"},
        channel_first=False,
        reverse_color=False,
        version: Version = None,
    ) -> None:
        """
        Args:
            path: The folder path containing images
            load_args: Args fed to:
                https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.open
            save_args:
                Args including:
                `suffix`: File suffix such as ".jpg"
                upper: If provided, used as the upper pixel value corresponding to 0xFF (255)
                    for linear scaling to ensure the pixel value is between 0 and 255.
                lower: If provided, used as the lower pixel value corresponding to 0x00 (0)
                    for linear scaling to ensure the pixel value is between 0 and 255.
                `mode` fed to:
                https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.fromarray
                Other args fed to:
                https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save

            channel_first: If true, the first dimension of 3-D array is
                treated as channel (color) as in PyTorch.
                If false, the last dimenstion of the 3-D array is
                treated as channel (color) as in TensorFlow, Pillow, and OpenCV.
            reverse_color: If true, the order of channel (color) is reversed
                (RGB to BGR when loading, BGR to RGB when saving).
                Set true to use packages wuch as OpenCV which uses BGR order natively.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
        """

        super().__init__(
            filepath=Path(path), version=version, exists_function=self._exists,
        )
        self._load_args = load_args
        self._save_args = save_args
        self._channel_first = channel_first
        self._reverse_color = reverse_color

    def _load(self) -> Any:

        load_path = Path(self._get_load_path())

        load_args = copy.deepcopy(self._load_args)
        load_args = load_args or dict()

        dict_structure = load_args.pop("dict_structure", True)
        as_numpy = load_args.pop("as_numpy", True)

        channel_first = self._channel_first
        reverse_color = self._reverse_color

        if load_path.is_dir():
            images_dict = {}
            for p in load_path.glob("*"):
                img = load_image(
                    p,
                    load_args,
                    as_numpy=as_numpy,
                    channel_first=channel_first,
                    reverse_color=reverse_color,
                )
                images_dict[p.stem] = img

            if dict_structure is None:
                return list(images_dict.values())
            if dict_structure == "sep_names":
                return dict(
                    images=list(images_dict.values()), names=list(images_dict.keys())
                )

            return images_dict

        else:
            return load_image(
                load_path,
                load_args,
                as_numpy=self.as_numpy,
                channel_first=channel_first,
                reverse_color=reverse_color,
            )

    def _save(self, data: Union[dict, list, np.ndarray, type(Image.Image)]) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)
        p = save_path

        save_args = copy.deepcopy(self._save_args)
        save_args = save_args or dict()
        suffix = save_args.pop("suffix", ".jpg")
        mode = save_args.pop("mode", None)
        upper = save_args.pop("upper", None)
        lower = save_args.pop("lower", None)
        to_scale = (upper is not None) or (lower is not None)

        if isinstance(data, dict):
            images = list(data.values())
            names = list(data.keys())
            if "names" in names and "images" in names:
                images = data.get("images")
                names = data.get("names")

        else:
            images = data
            names = None

        if hasattr(images, "save"):
            if not to_scale:
                img = images
                img.save(p, **save_args)
                return None
            else:
                images = np.asarray(images)

        if isinstance(images, np.ndarray):
            if self._channel_first:
                images = to_channel_last_arr(images)
            if self._reverse_color:
                images = ReverseChannel(channel_first=self._channel_first)(images)
            if images.ndim in {2, 3}:
                img = images
                img = scale(lower=lower, upper=upper)(img)
                img = np.squeeze(img)
                img = Image.fromarray(img, mode=mode)
                img.save(p, **save_args)
                return None
            elif images.ndim in {4}:
                images = scale(lower=lower, upper=upper)(images)
                dataset = Np3DArrDataset(images)
            else:
                raise ValueError(
                    "Unsupported number of dimensions: {}".format(images.ndim)
                )
        elif hasattr(images, "__getitem__") and hasattr(images, "__len__"):
            if not to_scale:
                p.mkdir(parents=True, exist_ok=True)
                for i, img in enumerate(images):
                    if isinstance(img, np.ndarray):
                        if self._channel_first:
                            img = to_channel_last_arr(img)
                        if self._reverse_color:
                            img = ReverseChannel(channel_first=self._channel_first)(img)
                        img = np.squeeze(img)
                        img = Image.fromarray(img)
                    name = names[i] if names else "{:05d}".format(i)
                    s = p / "{}{}".format(name, suffix)
                    img.save(s, **save_args)
                return None
            else:
                dataset = Np3DArrDatasetFromList(
                    images, transform=scale(lower=lower, upper=upper)
                )
        else:
            raise ValueError("Unsupported data type: {}".format(type(images)))

        p.mkdir(parents=True, exist_ok=True)
        for i in range(len(dataset)):
            img = dataset[i]
            if isinstance(img, (tuple, list)):
                img = img[0]
            if self._channel_first:
                img = to_channel_last_arr(img)
            if self._reverse_color:
                img = ReverseChannel(channel_first=self._channel_first)(img)
            img = np.squeeze(img)
            img = Image.fromarray(img, mode=mode)
            name = names[i] if names else "{:05d}".format(i)
            s = p / "{}{}".format(name, suffix)
            img.save(s, **save_args)
        return None

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            load_args=self._save_args,
            save_args=self._save_args,
            channel_first=self._channel_first,
            reverse_color=self._reverse_color,
            version=self._version,
        )

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).exists()


def load_image(
    load_path, load_args, as_numpy=False, channel_first=False, reverse_color=False
):
    with load_path.open("rb") as local_file:
        img = Image.open(local_file, **load_args)
        if as_numpy:
            img = np.asarray(img)
            if channel_first:
                img = to_channel_first_arr(img)
            if reverse_color:
                img = ReverseChannel(channel_first=channel_first)(img)
        return img


def scale(**kwargs):
    def _scale(a):
        lower = kwargs.get("lower")
        upper = kwargs.get("upper")
        if (lower is not None) or (upper is not None):
            max_val = a.max()
            min_val = a.min()
            stat_dict = dict(max_val=max_val, min_val=min_val)
            log.info(stat_dict)
            upper = upper or max_val
            lower = lower or min_val
            a = (
                ((a - min_val) / (max_val - min_val)) * (upper - lower) + lower
            ).astype(np.uint8)
        return a

    return _scale


class Np3DArrDataset:
    def __init__(self, a):
        self.a = a

    def __getitem__(self, index):
        return self.a[index, ...]

    def __len__(self):
        return len(self.a)


class Np3DArrDatasetFromList:
    def __init__(self, a, transform=None):
        self.a = a
        self.transform = transform

    def __getitem__(self, index):
        item = np.asarray(self.a[index])
        if self.transform:
            item = self.transform(item)
        return item

    def __len__(self):
        return len(self.a)



import copy
from pathlib import Path
from typing import Any, Dict
import logging

import cv2
from ..core import AbstractVersionedDataSet, DataSetError, Version

log = logging.getLogger(__name__)


class OpenCVImagesLocalDataSet(AbstractVersionedDataSet):
    def __init__(
        self,
        filepath: str,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
        channel_first=False,
        version: Version = None,
    ) -> None:

        super().__init__(
            filepath=Path(filepath), version=version, exists_function=self._exists,
        )
        self._load_args = load_args
        self._save_args = save_args
        self.channel_first = channel_first

    def _load(self) -> Any:

        load_path = Path(self._get_load_path())
        load_args = copy.deepcopy(self._load_args) or dict()
        img = cv2.imread(str(load_path), **load_args)

        return img

    def _save(self, img) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)
        cv2.imwrite(str(save_path), img)

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            load_args=self._save_args,
            save_args=self._save_args,
            version=self._version,
        )

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).exists()



from kedro.pipeline import Pipeline

from .context import KedroContext


class SavePipelineJsonContext(KedroContext):
    _pipeline_json_text_dataset = None

    def _save_pipeline_json(self, pipeline: Pipeline):
        if self._pipeline_json_text_dataset is not None:
            pipeline_json_str = pipeline.to_json()
            self._pipeline_json_text_dataset.save(pipeline_json_str)



import time

import logging

log = logging.getLogger(__name__)

__all__ = ["TimeLimit"]


class TimeLimit:
    def __init__(self, limit_sec=3600):
        """Time limit for training.

        Args:
            limit_sec (int, optional): Time limit in seconds. Defaults to 3600.
        """
        self.limit_sec = limit_sec
        self.start_time = time.time()

    def __call__(self, engine):
        elapsed_time = time.time() - self.start_time
        if elapsed_time > self.limit_sec:
            log.warning(
                "Reached the time limit: {} sec. Stop training".format(self.limit_sec)
            )
            engine.terminate()



import copy
import pandas as pd
import numpy as np
import logging

log = logging.getLogger(__name__)

FLOAT_TYPES = ["float16", "float32", "float64"]
INT_TYPES = ["int8", "int16", "int32", "int64"]


class DfBaseMethod:
    method = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs

        if self.method is None:
            mutated_df = df
        else:
            mutated_df = getattr(df, self.method)(**kwargs)
        return mutated_df


def _groupby(df, groupby, columns):
    if isinstance(columns, dict):
        df = df.rename(columns=columns)
        columns = list(columns.values())
    if groupby is not None:
        if not isinstance(groupby, dict):
            groupby = dict(by=groupby)
        df = df.groupby(**groupby)
    if columns is not None:
        if isinstance(columns, (list, str)):
            df = df[columns]
        else:
            raise ValueError("'{}' must be dict, list, or str.".format(columns))
    return df


def _add_mutated(df, mutated_df, columns, keep_others):
    if keep_others:
        columns_list = list(columns.values()) if isinstance(columns, dict) else columns
        df[columns_list] = mutated_df
    else:
        df = mutated_df
    return df


def _preprocess_columns(columns):
    assert isinstance(columns, (dict, list, str, type(None)))
    columns_dict = None
    if isinstance(columns, dict):
        columns_dict = copy.deepcopy(columns)
        columns = list(columns_dict.values())
    return columns_dict, columns


class DfBaseTask:
    method = None

    def __init__(
        self, groupby=None, columns=None, keep_others=False, method=None, **kwargs
    ):
        if self.method is None:
            self.method = method
        columns_dict, columns = _preprocess_columns(columns)
        self.groupby = groupby
        self.columns_dict = columns_dict
        self.columns = columns
        self.keep_others = keep_others
        self.kwargs = kwargs

    def __call__(self, df):
        groupby = self.groupby
        columns_dict = self.columns_dict
        columns = self.columns
        keep_others = self.keep_others
        kwargs = self.kwargs
        df = DfDuplicate(columns=columns_dict)(df)
        g_df = _groupby(df, groupby, columns)
        if self.method is None:
            mutated_df = g_df
        else:
            mutated_df = getattr(g_df, self.method)(**kwargs)
        return _add_mutated(df, mutated_df, columns, keep_others)


class DfResetIndex(DfBaseMethod):
    method = "reset_index"


class DfSetIndex(DfBaseMethod):
    method = "set_index"


class DfSortValues(DfBaseMethod):
    method = "sort_values"


class DfDrop(DfBaseMethod):
    method = "drop"


class DfQuery(DfBaseMethod):
    method = "query"


class DfDropDuplicates(DfBaseMethod):
    method = "drop_duplicates"


class DfGroupby(DfBaseMethod):
    method = "groupby"


class DfSelectDtypes(DfBaseMethod):
    method = "select_dtypes"


class DfTransform(DfBaseTask):
    method = "transform"


class DfApply(DfBaseTask):
    method = "apply"


class DfApplymap(DfBaseTask):
    method = "applymap"


class DfPipe(DfBaseTask):
    method = "pipe"


class DfAgg(DfBaseTask):
    method = "agg"


class DfAggregate(DfBaseTask):
    method = "aggregate"


class DfRolling(DfBaseTask):
    method = "rolling"


class DfExpanding(DfBaseTask):
    method = "expanding"


class DfEwm(DfBaseTask):
    method = "ewm"


class DfFillna(DfBaseTask):
    method = "fillna"


class DfHead(DfBaseTask):
    method = "head"


class DfTail(DfBaseTask):
    method = "tail"


class DfShift(DfBaseTask):
    method = "shift"


class DfResample(DfBaseTask):
    method = "resample"


class DfNgroup(DfBaseTask):
    method = "ngroup"


class DfMerge:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, left_df, right_df):
        kwargs = self.kwargs
        kwargs.setdefault("suffixes", (False, False))
        return pd.merge(left_df, right_df, **kwargs)


class DfConcat:
    def __init__(self, new_col_name=None, new_col_values=None, col_id=None, sort=False):
        assert isinstance(new_col_name, (str, type(None)))
        assert isinstance(new_col_values, (list, type(None)))
        assert isinstance(col_id, (str, type(None)))
        kwargs = dict(
            new_col_name=new_col_name,
            new_col_values=new_col_values,
            col_id=col_id,
            sort=sort,
        )
        self.kwargs = kwargs

    def __call__(self, *args):
        kwargs = self.kwargs
        df_list = [arg for arg in args if isinstance(arg, pd.DataFrame)]
        assert len(df_list) >= 1, "No data frame was fed."

        new_col_name = kwargs.get("new_col_name")
        new_col_values = kwargs.get("new_col_values")
        col_id = kwargs.get("col_id")
        sort = kwargs.get("sort", False)

        if col_id is not None:
            for df in df_list:
                df.set_index(keys=col_id, inplace=True)
        else:
            col_id = df_list[0].index.name

        if (new_col_name is not None) and (new_col_values is None):
            new_col_values = list(range(len(df_list)))

        names = [new_col_name, col_id] if new_col_name else col_id
        df = pd.concat(
            df_list,
            sort=sort,
            verify_integrity=bool(col_id),
            keys=new_col_values,
            names=names,
        )
        if new_col_name:
            df.reset_index(inplace=True, level=new_col_name)
        return df


class DfSample:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        frac = kwargs.get("frac", 1.0)
        random_state = kwargs.get("random_state")
        col_sample = kwargs.get("col_sample")
        col_assign = kwargs.get("col_assign")

        log.info("DF shape before random sampling: {}".format(df.shape))
        if col_sample:
            population_arr = df[col_sample].unique()
            size = int(len(population_arr) * frac)
            np.random.seed(random_state)
            sample_arr = np.random.choice(population_arr, size=size, replace=False)
            if col_assign:
                assert isinstance(col_assign, str)
                sample_set = set(sample_arr.tolist())
                df[col_assign] = df[col_sample].map(
                    lambda v: np.int8(v in sample_set), na_action="ignore"
                )
            else:
                sample_series = pd.Series(sample_arr, name=col_sample)
                df = pd.merge(df, sample_series, how="right", on=col_sample)
        if not col_sample:
            if col_assign:
                raise NotImplementedError("Specify col_sample to use col_assign.")
            df = df.sample(frac=frac, random_state=random_state)
        log.info("DF shape after random sampling: {}".format(df.shape))
        return df


class DfGetCols:
    def __call__(self, df):
        return df.columns.to_list()


class DfSelectDtypesCols(DfSelectDtypes):
    def __call__(self, df):
        return super().__call__(df).columns.to_list()


class DfGetColIndexes:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        cols = kwargs.get("cols")
        assert cols
        for col in cols:
            if col not in df.columns:
                log.warning("Could not find column: {}".format(col))
        cols = [col for col in cols if col in df.columns.to_list()]
        indices = [df.columns.to_list().index(col) for col in cols]
        return indices


class DfDropFilter:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        cols_drop = DfFilter(**kwargs)(df).columns.to_list()
        return df.drop(columns=cols_drop)


class DfAddRowStat:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        regex = kwargs.get("regex", r".*")
        prefix = kwargs.get("prefix", "stat_all")

        prefix = prefix or regex
        t_df = df.filter(regex=regex, axis=1)
        cols_float = t_df.select_dtypes(include=FLOAT_TYPES).columns.to_list()
        cols_int = t_df.select_dtypes(include=INT_TYPES).columns.to_list()

        if cols_float:
            df["{}_float_na_".format(prefix)] = (
                df[cols_float].isna().astype("int8").sum(axis=1)
            )

            df["{}_float_zero_".format(prefix)] = (
                (df[cols_float] == 0.0).astype("int8").sum(axis=1)
            )
            df["{}_float_pos_".format(prefix)] = (
                (df[cols_float] > 0.0).astype("int8").sum(axis=1)
            )
            df["{}_float_neg_".format(prefix)] = (
                (df[cols_float] < 0.0).astype("int8").sum(axis=1)
            )
            df["{}_float_pos_ones_".format(prefix)] = (
                (df[cols_float] == 1.0).astype("int8").sum(axis=1)
            )
            df["{}_float_neg_ones_".format(prefix)] = (
                (df[cols_float] == -1.0).astype("int8").sum(axis=1)
            )
            df["{}_float_gt_pos_ones_".format(prefix)] = (
                (df[cols_float] > 1.0).astype("int8").sum(axis=1)
            )
            df["{}_float_lt_neg_ones_".format(prefix)] = (
                (df[cols_float] < -1.0).astype("int8").sum(axis=1)
            )
            df["{}_float_max_".format(prefix)] = df[cols_float].max(axis=1)
            df["{}_float_min_".format(prefix)] = df[cols_float].min(axis=1)
            df["{}_float_mean_".format(prefix)] = df[cols_float].mean(axis=1)

        if cols_int:
            df["{}_int_zero_".format(prefix)] = (
                (df[cols_int] == 0).astype("int8").sum(axis=1)
            )
            df["{}_int_pos_".format(prefix)] = (
                (df[cols_int] > 0).astype("int8").sum(axis=1)
            )
            df["{}_int_neg_".format(prefix)] = (
                (df[cols_int] < 0).astype("int8").sum(axis=1)
            )
            df["{}_int_pos_ones_".format(prefix)] = (
                (df[cols_int] == 1).astype("int8").sum(axis=1)
            )
            df["{}_int_neg_ones_".format(prefix)] = (
                (df[cols_int] == -1).astype("int8").sum(axis=1)
            )
            df["{}_int_gt_pos_ones_".format(prefix)] = (
                (df[cols_int] > 1).astype("int8").sum(axis=1)
            )
            df["{}_int_lt_neg_ones_".format(prefix)] = (
                (df[cols_int] < -1).astype("int8").sum(axis=1)
            )
            df["{}_int_max_".format(prefix)] = df[cols_int].max(axis=1)
            df["{}_int_min_".format(prefix)] = df[cols_int].min(axis=1)
            df["{}_int_mean_".format(prefix)] = df[cols_int].mean(axis=1)

        return df


class DfEval:
    def __init__(self, expr, parser="pandas", engine=None, truediv=True):
        kwargs = dict(expr=expr, parser=parser, engine=engine, truediv=truediv)
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        return df.eval(**kwargs)


class DfFilter(DfBaseTask):
    def __call__(self, df):
        df = super().__call__(df)
        kwargs = self.kwargs
        items = kwargs.get("items")
        if items:
            if isinstance(items, str):
                items = [items]
            assert isinstance(items, list), "'items' should be a list or string"
            missing = [item for item in items if item not in df.columns]
            if missing:
                log.warning("filter could not find columns: {}".format(missing))
            items = [item for item in items if item in df.columns]
            kwargs.update(dict(items=items))
        return df.filter(**kwargs)


class DfFilterCols(DfFilter):
    def __call__(self, df):
        return super().__call__(df).columns.to_list()


class DfCondReplace:
    def __init__(self, flag, columns, value=np.nan, replace_if_flag=True, **kwargs):

        if not isinstance(flag, dict):
            flag = dict(expr=flag)
        columns_dict, columns = _preprocess_columns(columns)
        self.flag = flag
        self.columns_dict = columns_dict
        self.columns = columns
        self.value = value
        self.replace_if_flag = replace_if_flag
        self.kwargs = kwargs

    def __call__(self, df):
        flag = self.flag
        columns_dict = self.columns_dict
        columns = self.columns
        value = self.value
        replace_if_flag = self.replace_if_flag
        kwargs = self.kwargs

        df = DfDuplicate(columns=columns_dict)(df)
        cond = df.eval(**flag)
        if not replace_if_flag:
            cond = np.invert(cond)
        df.loc[cond, columns] = value
        return df


class DfRename:
    def __init__(self, index=None, columns=None, copy=True, level=None):
        kwargs = dict(index=index, columns=columns, copy=copy, level=level)
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        index = kwargs.get("index")
        columns = kwargs.get("columns")
        level = kwargs.get("level")
        if not (index is None and columns is None and level is None):
            df = df.rename(**kwargs)
        return df


class DfDuplicate:
    def __init__(self, columns):
        assert isinstance(columns, (dict, type(None)))
        columns_dict, columns = _preprocess_columns(columns)
        self.columns_dict = columns_dict

    def __call__(self, df):
        columns_dict = self.columns_dict

        if isinstance(columns_dict, dict):
            new_df = DfRename(columns=columns_dict)(df[columns_dict.keys()])
            df = pd.concat([df, new_df], axis=1, sort=False)
        return df


class DfMap:
    def __init__(self, arg, prefix="", suffix="", **kwargs):
        assert isinstance(arg, dict)
        self.arg = arg
        self.prefix = prefix
        self.suffix = suffix
        self.kwargs = kwargs

    def __call__(self, df):
        arg = self.arg
        prefix = self.prefix
        suffix = self.suffix
        kwargs = self.kwargs
        for col, m in arg.items():
            if col in df.columns:
                new_col = prefix + col + suffix
                if isinstance(m, pd.Series):
                    m = m.to_dict()
                if isinstance(m, dict):
                    df.loc[:, new_col] = df[col].map(m, **kwargs)
                elif callable(m):
                    df.loc[:, new_col] = df[col].apply(m, **kwargs)
                elif m is not None:
                    df.loc[:, new_col] = m
            else:
                log.warning("'{}' not in the DataFrame".format(col))
        return df


class SrMap:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, sr):
        kwargs = self.kwargs
        return sr.map(**kwargs)


class DfGetDummies:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        return pd.get_dummies(df, **kwargs)


class DfColApply:
    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs

    def __call__(self, df):
        func = self.func
        kwargs = self.kwargs
        func_params = kwargs.pop("func_params", dict())
        cols = kwargs.pop("cols", None)
        cols = cols or df.columns
        df.loc[:, cols] = func(df.loc[:, cols], **func_params)
        return df


class DfDtypesApply:
    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs

    def __call__(self, df):
        func = self.func
        kwargs = self.kwargs
        func_params = kwargs.pop("func_params", dict())
        cols = df.select_dtypes(**kwargs).columns.to_list()
        df.loc[:, cols] = func(df.loc[:, cols], **func_params)
        return df


class DfRowApply:
    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs

    def __call__(self, df):
        func = self.func
        kwargs = self.kwargs
        func_params = kwargs.pop("func_params", dict())
        rows = kwargs.pop("rows", None)
        rows = rows or df.index
        df.loc[rows, :] = func(df.loc[rows, :], **func_params)
        return df


def _cols_apply(df, func, cols, kwargs):
    assert callable(func)
    assert isinstance(kwargs, dict)
    if isinstance(cols, tuple):
        cols = list(cols)
    if not isinstance(cols, list):
        cols = [cols]
    for col in cols:
        assert isinstance(col, str), "'{}' is not str.".format(col)
        if col in df.columns:
            df.loc[:, col] = func(df.loc[:, col], **kwargs)
        else:
            log.warning("'{}' not in the data frame.".format(col))
    return df


class DfToDatetime:
    def __init__(self, cols=None, **kwargs):
        self.cols = cols
        self.kwargs = kwargs

    def __call__(self, df):
        cols = self.cols
        kwargs = self.kwargs
        if cols:
            df = _cols_apply(df, func=pd.to_datetime, cols=cols, kwargs=kwargs)
        else:
            df = pd.to_datetime(df, **kwargs)
        return df


class DfTotalSeconds:
    def __init__(self, cols, **kwargs):
        self.cols = cols
        self.kwargs = kwargs

    def __call__(self, df):
        cols = self.cols
        kwargs = self.kwargs
        df = _cols_apply(df, func=pd.Series.dt.total_seconds, cols=cols, kwargs=kwargs)
        return df


class DfToTimedelta:
    def __init__(self, cols, **kwargs):
        self.cols = cols
        self.kwargs = kwargs

    def __call__(self, df):
        cols = self.cols
        kwargs = self.kwargs
        df = _cols_apply(df, func=pd.to_timedelta, cols=cols, kwargs=kwargs)
        return df


class DfStrftime:
    def __init__(self, cols, **kwargs):
        kwargs.setdefault("date_format", "%Y-%m-%dT%H:%M:%S")
        self.cols = cols
        self.kwargs = kwargs

    def __call__(self, df):
        cols = self.cols
        kwargs = self.kwargs
        df = _cols_apply(df, func=pd.Series.dt.strftime, cols=cols, kwargs=kwargs)
        return df


class DfSlice:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        start = kwargs.get("start", 0)
        end = kwargs.get("end", df.shape[0])
        step = kwargs.get("step", 1)
        return df.loc[start:end:step, :]


class DfFocusTransform:
    def __init__(
        self, focus, columns, groupby=None, keep_others=False, func="max", **kwargs
    ):
        assert isinstance(focus, (dict, str))
        assert isinstance(columns, (dict, list, str))
        columns_dict, columns = _preprocess_columns(columns)
        self.focus = focus
        self.columns_dict = columns_dict
        self.columns = columns
        self.groupby = groupby
        self.keep_others = keep_others
        self.func = func
        self.kwargs = kwargs

    def __call__(self, df):
        focus = self.focus
        columns_dict = self.columns_dict
        columns = self.columns
        groupby = self.groupby
        keep_others = self.keep_others
        func = self.func
        kwargs = self.kwargs
        df = DfCondReplace(
            replace_if_flag=False, flag=focus, columns=columns_dict, value=np.nan
        )(df)
        df = DfTransform(
            groupby=groupby,
            columns=columns,
            keep_others=keep_others,
            func=func,
            **kwargs
        )(df)
        return df


class DfRelative:
    def __init__(self, focus, columns, groupby=None):
        assert isinstance(focus, (dict, str))
        assert isinstance(columns, dict)
        self.focus = focus
        self.columns = columns
        self.groupby = groupby

    def __call__(self, df):
        focus = self.focus
        columns = self.columns
        groupby = self.groupby
        df = DfFocusTransform(focus=focus, columns=columns, groupby=groupby)(df)
        for col, new_col in columns.items():
            df[new_col] = df[col] - df[new_col]
        return df


def distance_matrix(coo_2darr, ord=None):
    coo_3darr = np.expand_dims(coo_2darr, axis=1)
    dist_2darr = np.linalg.norm(coo_3darr - coo_2darr, axis=2, ord=ord)
    return dist_2darr


def distance_to_affinity(
    dist_2darr,
    unit_distance=1.0,
    affinity_scale=1.0,
    binary_affinity=False,
    min_affinity=1.0e-6,
):
    if binary_affinity:
        affinity_2darr = dist_2darr <= unit_distance
    else:
        affinity_2darr = np.exp(-0.5 * np.square(dist_2darr / unit_distance))
    if affinity_scale is not None:
        affinity_2darr = affinity_2darr * affinity_scale
    if min_affinity is not None:
        affinity_2darr[affinity_2darr < min_affinity] = 0
    return affinity_2darr


def affinity_matrix(
    coo_2darr,
    ord=None,
    unit_distance=1.0,
    affinity_scale=1.0,
    binary_affinity=False,
    min_affinity=1.0e-6,
    zero_diag=True,
):
    dist_2darr = distance_matrix(coo_2darr, ord=ord)
    affinity_2darr = distance_to_affinity(
        dist_2darr,
        unit_distance=unit_distance,
        affinity_scale=affinity_scale,
        binary_affinity=binary_affinity,
        min_affinity=min_affinity,
    )
    if zero_diag:
        size = affinity_2darr.shape[1]
        affinity_2darr = affinity_2darr * (np.ones((size, size)) - np.eye(size))
    return affinity_2darr


def degree_matrix(affinity_2darr):
    return np.diagflat(affinity_2darr.sum(axis=1))


def laplacian_matrix(
    coo_2darr,
    ord=None,
    unit_distance=1.0,
    affinity_scale=1.0,
    binary_affinity=False,
    min_affinity=1.0e-6,
):
    affinity_2darr = affinity_matrix(
        coo_2darr=coo_2darr,
        ord=ord,
        unit_distance=unit_distance,
        affinity_scale=affinity_scale,
        binary_affinity=binary_affinity,
        min_affinity=min_affinity,
        zero_diag=True,
    )
    laplacian_2darr = degree_matrix(affinity_2darr) - affinity_2darr
    return laplacian_2darr


def row_vector_to_square_matrix(a):
    return np.tile(a, (a.shape[0], 1))


def eigen(
    a,
    return_values=True,
    values_as_square_matrix=False,
    return_vectors=False,
    sort=False,
):
    if return_vectors:
        w, v = np.linalg.eigh(a)

        if v.dtype == np.complex128:
            log.warning("Complex eigenvectors. The imaginary parts are discarded.")
            v = np.real(v)

        if sort:
            idx = np.argsort(w)
            v = v[:, idx]

        if return_values:
            w = w[idx]
            if values_as_square_matrix:
                w = row_vector_to_square_matrix(w)
            return w, v

        return v

    else:
        w = np.linalg.eigvalsh(a)

        if sort:
            idx = np.argsort(w)
            w = w[idx]

        if values_as_square_matrix:
            w = row_vector_to_square_matrix(w)

        return w


def laplacian_eigen(
    coo_2darr,
    return_values=True,
    return_vectors=False,
    ord=None,
    unit_distance=1.0,
    affinity_scale=1.0,
    binary_affinity=False,
    min_affinity=1.0e-6,
    sort=False,
):
    a = laplacian_matrix(
        coo_2darr,
        ord=ord,
        unit_distance=unit_distance,
        affinity_scale=affinity_scale,
        binary_affinity=binary_affinity,
        min_affinity=min_affinity,
    )
    ev = eigen(
        a,
        return_values=return_values,
        values_as_square_matrix=True,
        return_vectors=return_vectors,
        sort=sort,
    )
    return ev


class DfAssignColumns:
    def __init__(self, names=None, name_fmt="{:03d}"):
        if names is None:
            assert isinstance(name_fmt, str)
        self.names = names
        self.name_fmt = name_fmt

    def __call__(self, df, values):
        names = self.names
        name_fmt = self.name_fmt
        names = names or [name_fmt.format(i) for i in range(values.shape[1])]
        values_df = pd.DataFrame(values, columns=names)
        out_df = pd.concat([df.reset_index(drop=True), values_df], axis=1, sort=False)
        return out_df.set_index(df.index)


class DfSpatialFeatures:
    def __init__(
        self,
        output="distance",
        coo_cols=["X", "Y"],
        groupby=None,
        ord=None,
        unit_distance=1.0,
        affinity_scale=1.0,
        binary_affinity=False,
        min_affinity=1.0e-6,
        col_name_fmt="feat_{:03d}",
        keep_others=True,
        sort=True,
    ):
        """
        Available values for output:
         distance
         affinity
         laplacian
         eigenvalues
         eigenvectors
         n_connected
        """

        kwargs = dict(
            output=output,
            coo_cols=coo_cols,
            groupby=groupby,
            ord=ord,
            unit_distance=unit_distance,
            affinity_scale=affinity_scale,
            binary_affinity=binary_affinity,
            min_affinity=min_affinity,
            col_name_fmt=col_name_fmt,
            keep_others=keep_others,
            sort=sort,
        )
        self.kwargs = kwargs

    def __call__(self, df):
        kwargs = self.kwargs
        output = kwargs.get("output")
        coo_cols = kwargs.get("coo_cols")
        groupby = kwargs.get("groupby")
        ord = kwargs.get("ord")
        unit_distance = kwargs.get("unit_distance")
        affinity_scale = kwargs.get("affinity_scale")
        binary_affinity = kwargs.get("binary_affinity")
        min_affinity = kwargs.get("min_affinity")
        col_name_fmt = kwargs.get("col_name_fmt")
        keep_others = kwargs.get("keep_others")
        sort = kwargs.get("sort")

        for coo_col in coo_cols:
            assert coo_col in df.columns

        if isinstance(affinity_scale, str):
            affinity_scale = dict(expr=affinity_scale)

        if groupby is not None:
            if not isinstance(groupby, dict):
                groupby = dict(by=groupby)
            g_df_iter = df.groupby(**groupby)
        else:
            g_df_iter = [("", df)]

        output_2darr_list = []
        for g_name, g_df in g_df_iter:
            if isinstance(affinity_scale, dict):
                affinity_scale_1darr = g_df.eval(**affinity_scale).values
                affinity_scale = np.expand_dims(
                    affinity_scale_1darr, axis=0
                ) * np.expand_dims(affinity_scale_1darr, axis=1)

            coo_2darr = g_df[coo_cols].values
            if output in ["distance"]:
                output_2darr = distance_matrix(coo_2darr, ord=ord)
            elif output in ["affinity"]:
                output_2darr = affinity_matrix(
                    coo_2darr=coo_2darr,
                    ord=ord,
                    unit_distance=unit_distance,
                    affinity_scale=affinity_scale,
                    binary_affinity=binary_affinity,
                    min_affinity=min_affinity,
                    zero_diag=True,
                )
            elif output in ["laplacian"]:
                output_2darr = laplacian_matrix(
                    coo_2darr,
                    ord=ord,
                    unit_distance=unit_distance,
                    affinity_scale=affinity_scale,
                    binary_affinity=binary_affinity,
                    min_affinity=min_affinity,
                )
            elif output in ["eigenvalues", "eigenvectors", "n_connected"]:
                if output in ["eigenvectors"]:
                    return_values = False
                    return_vectors = True
                else:
                    return_values = True
                    return_vectors = False

                output_2darr = laplacian_eigen(
                    coo_2darr,
                    return_values=return_values,
                    return_vectors=return_vectors,
                    ord=ord,
                    unit_distance=unit_distance,
                    affinity_scale=affinity_scale,
                    binary_affinity=binary_affinity,
                    min_affinity=min_affinity,
                    sort=sort,
                )
                if output in ["n_connected"]:
                    output_2darr = np.sum(
                        (output_2darr < 1.0e-8).astype(np.uint64), axis=1, keepdims=True
                    )
            else:
                raise NotImplementedError(output)
            output_2darr_list.append(output_2darr)

        output_2darr = np.concatenate(output_2darr_list, axis=0)

        if col_name_fmt is None:
            return output_2darr

        else:
            assert isinstance(col_name_fmt, str)
            if keep_others:
                return DfAssignColumns(name_fmt=col_name_fmt)(df, output_2darr)
            else:
                output_col_names = [
                    col_name_fmt.format(i) for i in range(output_2darr.shape[1])
                ]
                return pd.DataFrame(
                    output_2darr, columns=output_col_names, index=df.index
                )


def nested_dict_to_df(d, row_oriented=True, index_name="index", reset_index=True):
    df = pd.DataFrame(d)
    if row_oriented:
        df = df.transpose()
    if index_name is not None:
        df.index.name = index_name
    if reset_index:
        df.reset_index(inplace=True)
    return df


class NestedDictToDf:
    def __init__(self, row_oriented=True, index_name="index", reset_index=True):
        self.kwargs = dict(
            row_oriented=row_oriented, index_name=index_name, reset_index=reset_index
        )

    def __call__(self, d):
        return nested_dict_to_df(d, **self.kwargs)



import logging
from functools import wraps
from typing import Callable

from importlib.util import find_spec

if find_spec("pynvml"):
    from pynvml import (
        nvmlInit,
        nvmlSystemGetDriverVersion,
        nvmlSystemGetNVMLVersion,
        nvmlDeviceGetCount,
        nvmlDeviceGetHandleByIndex,
        nvmlDeviceGetName,
        nvmlDeviceGetMemoryInfo,
        nvmlDeviceGetUtilizationRates,
        nvmlShutdown,
    )
elif find_spec("py3nvml"):
    from py3nvml.py3nvml import (
        nvmlInit,
        nvmlSystemGetDriverVersion,
        nvmlSystemGetNVMLVersion,
        nvmlDeviceGetCount,
        nvmlDeviceGetHandleByIndex,
        nvmlDeviceGetName,
        nvmlDeviceGetMemoryInfo,
        nvmlDeviceGetUtilizationRates,
        nvmlShutdown,
    )
else:
    raise ImportError("Install pynvml or py3nvml.")


def _func_full_name(func: Callable):
    return getattr(func, "__qualname__", repr(func))


def get_nv_info():
    nv_info = dict()
    try:
        nvmlInit()

        nv_info["_Driver_Version"] = str(nvmlSystemGetDriverVersion(), errors="ignore")
        nv_info["_NVML_Version"] = str(nvmlSystemGetNVMLVersion(), errors="ignore")

        device_count = nvmlDeviceGetCount()
        nv_info["Device_Count"] = device_count

        devices = []

        for i in range(device_count):
            dev_info = dict()

            handle = nvmlDeviceGetHandleByIndex(i)
            dev_info["_Name"] = str(nvmlDeviceGetName(handle), errors="ignore")

            memory_info = nvmlDeviceGetMemoryInfo(handle)
            dev_info["Total_Memory"] = memory_info.total
            dev_info["Free_Memory"] = memory_info.free
            dev_info["Used_Memory"] = memory_info.used

            util_rates = nvmlDeviceGetUtilizationRates(handle)
            dev_info["GPU_Utilization_Rate"] = util_rates.gpu
            dev_info["Memory_Utilization_Rate"] = util_rates.memory

            devices.append(dev_info)

        nv_info["Devices"] = devices

        nvmlShutdown()

    except Exception as e:
        nv_info["Exception"] = str(e)

    return nv_info


def nvml_profile(func: Callable) -> Callable:
    @wraps(func)
    def _nvml_profile(*args, **kwargs):
        log = logging.getLogger(__name__)

        init_nv_info = get_nv_info()
        init_devices = init_nv_info.get("Devices", [])
        result = func(*args, **kwargs)
        nv_info = get_nv_info()
        devices = nv_info.get("Devices", [])
        device_count = nv_info.get("Device_Count", 0)

        used_memory_diffs = []
        for i in range(device_count):
            init_used_memory = init_devices[i].get("Used_Memory", 0)
            used_memory = devices[i].get("Used_Memory", 0)
            try:
                used_memory_diff = used_memory - init_used_memory
            except:
                used_memory_diff = None
            used_memory_diffs.append(used_memory_diff)

        log.info(
            "Ran: '{}', NVML returned: {}, Used memory diff: {}".format(
                _func_full_name(func), nv_info, used_memory_diffs,
            )
        )
        return result

    return _nvml_profile



import kedro
from .sub_pipeline import SubPipeline
from typing import Callable, Union, List, Iterable  # NOQA


class FlexiblePipeline(kedro.pipeline.Pipeline):
    def __init__(
        self,
        nodes,  # type: Iterable[Union[SubPipeline, kedro.pipeline.Pipeline, kedro.pipeline.node.Node]]
        *,
        parameters_in_inputs=False,  # type: bool
        module="",  # type: str
        decorator=[],  # type: Union[Callable, List[Callable]]
        **kwargs
    ):

        for i, node in enumerate(nodes):

            assert node is not None, "Node {}: is empty.".format(i)
            if isinstance(node, dict):
                assert (
                    "inputs" in node
                ), "Node {} ({}): is missing 'inputs' key.".format(i, node)
                assert (
                    "outputs" in node
                ), "Node {} ({}): is missing 'outputs' key.".format(i, node)

                if parameters_in_inputs:
                    inputs = node.get("inputs")
                    inputs = inputs if isinstance(inputs, list) else [inputs]
                    if not ("parameters" in inputs):
                        node["inputs"] = inputs + ["parameters"]

                node.setdefault("module", module)

                node.setdefault("decorator", [])
                if not isinstance(node["decorator"], list):
                    node["decorator"] = [node["decorator"]]

                decorator = decorator or []
                if not isinstance(decorator, list):
                    decorator = [decorator]

                node["decorator"] = decorator + node["decorator"]

                nodes[i] = SubPipeline(**node)

        super().__init__(nodes, **kwargs)



from typing import Union, Sequence

import torch

from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric
from ignite.metrics.metric import sync_all_reduce, reinit__is_reduced
from sklearn.metrics import cohen_kappa_score

__all__ = ["CohenKappaScore"]


class CohenKappaScore(Metric):
    """
    Calculates the cohen kappa score.
    - `update` must receive output of the form `(y_pred, y)` or `{'y_pred': y_pred, 'y': y}`.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        super(CohenKappaScore, self).__init__(device="cpu")

    @reinit__is_reduced
    def reset(self) -> None:
        self._sum_of_cohen_kappa = 0.0
        self._num_examples = 0

    @reinit__is_reduced
    def update(self, output: Sequence[torch.Tensor]) -> None:
        y_pred, y = output
        assert y_pred.shape == y.shape
        y_pred_arr = y_pred.numpy().flatten().astype(int)
        y_arr = y.numpy().flatten().astype(int)
        cohen_kappa = cohen_kappa_score(y_pred_arr, y_arr, *self.args, **self.kwargs)
        self._sum_of_cohen_kappa += cohen_kappa
        self._num_examples += y.shape[0]

    @sync_all_reduce("_sum_of_squared_errors", "_num_examples")
    def compute(self) -> Union[float, torch.Tensor]:
        if self._num_examples == 0:
            raise NotComputableError(
                "MeanSquaredError must have at least one example before it can be computed."
            )
        return self._sum_of_cohen_kappa / self._num_examples



# Copyright 2020 QuantumBlack Visual Analytics Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# NONINFRINGEMENT. IN NO EVENT WILL THE LICENSOR OR OTHER CONTRIBUTORS
# BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF, OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# The QuantumBlack Visual Analytics Limited ("QuantumBlack") name and logo
# (either separately or in combination, "QuantumBlack Trademarks") are
# trademarks of QuantumBlack. The License does not grant you any right or
# license to the QuantumBlack Trademarks. You may not use the QuantumBlack
# Trademarks or any confusingly similar mark as a trademark for your product,
#     or use the QuantumBlack Trademarks in any other manner that might cause
# confusion in the marketplace, including but not limited to in advertising,
# on websites, or on software.
#
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=no-member
import PIL
import io

import pytest

from pipelinex import AsyncAPIDataSet as APIDataSet
from kedro.io.core import DataSetError


def test_successfully_load_with_content_response():
    api_data_set = APIDataSet(
        url="https://raw.githubusercontent.com/quantumblacklabs/kedro/develop/img/kedro_banner.png",
        method="GET",
        attribute="content",
    )
    content = api_data_set.load()
    assert content[1:4] == b"PNG"  # part of PNG file signature


def test_successfully_load_with_response_itself():
    api_data_set = APIDataSet(
        url="https://raw.githubusercontent.com/quantumblacklabs/kedro/develop/img/kedro_banner.png",
        method="GET",
        attribute="",
    )
    response = api_data_set.load()
    content = response.content
    assert content[1:4] == b"PNG"  # part of PNG file signature


def test_attribute_not_found():
    attribute = "wrong_attribute"
    api_data_set = APIDataSet(
        url="https://raw.githubusercontent.com/quantumblacklabs/kedro/develop/img/kedro_banner.png",
        method="GET",
        attribute=attribute,
    )
    pattern = r"Response has no attribute: {}".format(attribute)

    with pytest.raises(DataSetError, match=pattern):
        api_data_set.load()


foobar_prefix = "https://raw.githubusercontent.com/"
foo_image_url = foobar_prefix + "quantumblacklabs/kedro/develop/img/kedro_banner.png"
bar_image_url = (
    foobar_prefix + "quantumblacklabs/kedro/develop/img/pipeline_visualisation.png"
)


def test_successfully_load_from_url_dict_with_content_response():
    api_data_set = APIDataSet(
        url={"foo_image.png": foo_image_url, "bar_image.png": bar_image_url},
        method="GET",
        attribute="content",
        pool_config={
            foobar_prefix: {
                "pool_connections": 1,
                "pool_maxsize": 1,
                "max_retries": 0,
                "pool_block": False,
            }
        },
    )
    content_dict = api_data_set.load()
    assert isinstance(content_dict, dict)
    for content in content_dict.values():
        assert content[1:4] == b"PNG"  # part of PNG file signature


def test_successfully_load_from_url_list_with_content_response():
    api_data_set = APIDataSet(
        url=[foo_image_url, bar_image_url],
        method="GET",
        attribute="content",
        pool_config={
            foobar_prefix: {
                "pool_connections": 1,
                "pool_maxsize": 1,
                "max_retries": 0,
                "pool_block": False,
            }
        },
    )
    content_list = api_data_set.load()
    assert isinstance(content_list, list)
    for content in content_list:
        assert content[1:4] == b"PNG"  # part of PNG file signature


def test_successfully_load_from_url_list_with_content_response_by_call():
    api_data_set = APIDataSet(
        pool_config={
            foobar_prefix: {
                "pool_connections": 1,
                "pool_maxsize": 1,
                "max_retries": 0,
                "pool_block": False,
            }
        },
    )
    content_list = api_data_set(
        url=[foo_image_url, bar_image_url], method="GET", attribute="content",
    )
    assert isinstance(content_list, list)
    for content in content_list:
        assert content[1:4] == b"PNG"  # part of PNG file signature


def test_successfully_load_from_url_list_with_content_response_by_call_twice():
    api_data_set = APIDataSet(
        pool_config={
            foobar_prefix: {
                "pool_connections": 1,
                "pool_maxsize": 1,
                "max_retries": 0,
                "pool_block": False,
            }
        },
    )
    content_list = api_data_set(
        url=[foo_image_url, bar_image_url], method="GET", attribute="content",
    )
    assert isinstance(content_list, list)
    for content in content_list:
        assert content[1:4] == b"PNG"  # part of PNG file signature

    content_list = api_data_set(
        url=[foo_image_url, bar_image_url], method="GET", attribute="content",
    )
    assert isinstance(content_list, list)
    for content in content_list:
        assert content[1:4] == b"PNG"  # part of PNG file signature


def test_successfully_load_from_url_list_with_transforms_by_call():
    api_data_set = APIDataSet(
        pool_config={
            foobar_prefix: {
                "pool_connections": 1,
                "pool_maxsize": 1,
                "max_retries": 0,
                "pool_block": False,
            }
        },
        transforms=[io.BytesIO, PIL.Image.open],
    )
    image_list = api_data_set(
        url=[foo_image_url, bar_image_url], method="GET", attribute="content",
    )
    for image in image_list:
        assert isinstance(image, PIL.Image.Image)



from pathlib import Path
from logging import getLogger

log = getLogger(__name__)

try:
    from kedro.framework.hooks import hook_impl
except ModuleNotFoundError:

    def hook_impl(func):
        return func


def mlflow_log_artifacts(paths, artifact_path=None, enable_mlflow=True):
    if enable_mlflow:
        paths = paths or []
        if not isinstance(paths, (list, tuple)):
            paths = [paths]

        try:
            from mlflow import log_artifact, log_artifacts

            for path in paths:
                if Path(path).is_file():
                    log_artifact(path, artifact_path)
                    log.info("File at '{}' was logged by MLflow.".format(path))
                elif Path(path).is_dir():
                    log_artifacts(path, artifact_path)
                    log.info("Directory at '{}' was logged by MLflow.".format(path))
                else:
                    log.warning("'{}' was not found.".format(path))
        except Exception:
            log.warning(
                "{} failed to be logged by MLflow.".format(paths), exc_info=True
            )


def mlflow_log_metrics(metrics, step=None, enable_mlflow=True):
    log.info("{}".format(metrics))

    if enable_mlflow:
        try:
            metrics = {k: float(v) for (k, v) in metrics.items()}

            from mlflow import log_metrics

            log_metrics(metrics, step)
        except Exception:
            log.warning(
                "{} failed to be logged by MLflow.".format(metrics), exc_info=True
            )


def mlflow_log_params(params, enable_mlflow=True):
    log.info("{}".format(params))

    if enable_mlflow:
        try:
            params = {
                k: ("{}".format(v)[:250] if v else "") for (k, v) in params.items()
            }

            from mlflow import log_params

            log_params(params)
        except Exception:
            log.warning(
                "{} failed to be logged by MLflow.".format(params), exc_info=True
            )


def mlflow_start_run(
    uri, experiment_name, artifact_location, run_name=None, enable_mlflow=True
):
    if enable_mlflow:

        from mlflow import (
            create_experiment,
            get_experiment_by_name,
            start_run,
            set_tracking_uri,
        )
        from mlflow.exceptions import MlflowException

        if uri:
            set_tracking_uri(uri)

        if experiment_name:
            try:
                experiment_id = create_experiment(
                    experiment_name, artifact_location=artifact_location,
                )
            except MlflowException:
                experiment_id = get_experiment_by_name(experiment_name).experiment_id
            start_run(experiment_id=experiment_id, run_name=run_name)


def mlflow_end_run(enable_mlflow=True):
    if enable_mlflow:
        try:
            from mlflow import end_run

            end_run()
        except Exception:
            log.warning("Failed to end MLflow run.", exc_info=True)



import sys
import os
from pathlib import Path
from typing import Union

from .context.context import KedroContext
from .context.flexible_context import FlexibleContext


def configure_source(project_path: Union[str, Path], source_dir="src"):

    if isinstance(project_path, str):
        project_path = Path(project_path)

    source_path = (project_path / source_dir).resolve()
    source_path_str = str(source_path)

    if source_path_str not in sys.path:
        sys.path.insert(0, source_path_str)

    if "PYTHONPATH" not in os.environ:
        os.environ["PYTHONPATH"] = source_path_str

    return source_path


def load_context(project_path, **kwargs) -> KedroContext:
    return FlexibleContext(project_path, **kwargs)



from importlib.util import find_spec

from .mlflow_utils import hook_impl, mlflow_log_metrics, mlflow_log_params


class MLflowOutputsLoggerHook:
    """ Logs output datasets of (list of) float/int and str classes to MLflow
    """

    def __init__(self, enable_mlflow: bool = True):
        """
        Args:
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow

    @hook_impl
    def after_node_run(self, node, catalog, inputs, outputs):

        for name, value in outputs.items():
            if isinstance(value, str):
                mlflow_log_params({name: value}, enable_mlflow=self.enable_mlflow)
                continue
            if isinstance(value, (float, int)):
                mlflow_log_metrics({name: value}, enable_mlflow=self.enable_mlflow)
                continue
            if isinstance(value, (list, tuple)):
                if all([isinstance(e, str) for e in value]):
                    mlflow_log_params(
                        {name: ", ".join(value)}, enable_mlflow=self.enable_mlflow
                    )
                    continue
                for i, e in enumerate(value):
                    if isinstance(e, (float, int)):
                        mlflow_log_metrics(
                            {name: e}, step=i, enable_mlflow=self.enable_mlflow
                        )
                    else:
                        break
